# 零、前言

TensorFlow 是用于机器学习以及最近的深度学习的最受欢迎的架构之一。 本书是将 TensorFlow 和 Keras 模型部署到实际应用中的指南。

本书从专门的蓝图开始，说明如何构建可生成预测的应用。 随后的每节课程都针对特定类型的模型，例如神经网络，使用 Keras 配置深度学习环境，并着重于三个重要问题，即模型如何工作，如何在示例模型中提高我们的预测准确率以及如何使用实际应用评估其表现。

在本书中，您将学习如何创建一个应用，该应用可以从深度学习中生成预测。 这项学习之旅从探索神经网络的通用组件及其基本表现开始。 在课程结束时，您将探索使用 TensorFlow 创建的训练有素的神经网络。 在其余的课程中，您将学习建立一个将不同组件组合在一起的深度学习模型，并评估它们在预测中的表现。 最后，我们将能够部署可运行的 Web 应用

到本书结尾，您将可以通过创建全新的模型并根据需要更改应用的核心组件来创建更准确的预测。

# 这本书涵盖的内容

第 1 课，“神经网络和深度学习简介”可帮助您设置和配置深度学习环境，并开始研究各个模型和案例研究。 它还讨论了神经网络及其思想及其起源，并探讨了其功能。

第 2 课，“模型架构”显示了如何使用深度学习模型预测比特币价格。

第 3 课，“模型评估和优化”展示了如何评估神经网络模型。 我们将修改网络的超参数以提高其表现。

第 4 课，“产品化”解释了如何将深度学习模型产品化，并提供了如何将模型部署为 Web 应用的练习。

# 这本书需要什么

本书将要求以下最低硬件要求：

*   处理器：1.8 GHz 或更高
*   内存：2 GB RAM
*   硬盘：10 GB

在本书中，我们将使用 Python 3，TensorFlow，TensorBoard 和 Keras。 请确保在计算机上安装了以下软件：

*   代码编辑器，例如：[Visual Studio Code](https://code.visualstudio.com/)
*   Python 3.6
*   Windows 上的 TensorFlow 1.4 或更高版本
*   Keras 2
*   TensorBoard
*   Jupyter 笔记本
*   Pandas
*   NumPy
*   操作系统：Windows（8 或更高版本），MacOS 或 Linux（Ubuntu）

# 这本书适合谁

本书专为对使用 TensorFlow 和 Keras 开发应用感兴趣的开发人员，分析师和数据科学家而设计。 您需要具备编程知识。 我们还假设您熟悉 Python 3 和 Web 应用的基本知识。 您还需要对线性代数，概率和统计数据有事先的了解和使用知识。

# 约定

在本书中，您将找到许多可以区分不同类型信息的文本样式。 以下是这些样式的一些示例，并解释了其含义。

文本，数据库表名称，文件夹名称，文件名，文件扩展名，路径名，虚拟 URL，用户输入和 Twitter 句柄中的代码字显示如下：“ `\`类提供了静态方法来生成自身的实例，例如`()`。”

代码块设置如下：

```py
tf.nn.max_pool(
activation,
ksize=[1, 2, 2, 1],
strides=[1, 2, 2, 1],
padding="SAME")
```

任何命令行输入或输出的编写方式如下：

```py
$ python3 lesson_1/activity_1/test_stack.py
```

**新术语**和**重要词**以粗体显示。 您在屏幕上看到的字词，例如在菜单或对话框中，将以如下形式显示在文本中：“单击“下一步”按钮将您移至下一个屏幕。”

### 注意

警告或重要提示会出现在这样的框中。

### 提示

提示和技巧如下所示。

# 读者反馈

始终欢迎读者的反馈。 让我们知道您对这本书的看法-您喜欢或不喜欢的东西。 读者反馈对我们很重要，因为它可以帮助我们开发出您真正能充分利用的标题。

要向我们发送一般反馈，只需发送电子邮件`<[feedback@packtpub.com](mailto:feedback@packtpub.com)>`，然后在您的邮件主题中提及该书的标题。

如果您有专业知识的主题，并且对写作或撰写书籍感兴趣，请参阅 [www.packtpub.com/authors](http://www.packtpub.com/authors) 上的作者指南。

# 客户支持

既然您是 Packt 书的骄傲拥有者，我们可以通过很多方法来帮助您从购买中获得最大收益。

# 下载示例代码

您可以从[这里](http://www.packtpub.com)的帐户中下载本书的示例代码文件。 如果您在其他地方购买了此书，则可以访问[这里](http://www.packtpub.com/support)并注册以将文件直接通过电子邮件发送给您。

您可以按照以下步骤下载代码文件：

1.  使用您的电子邮件地址和密码登录或注册到我们的网站。
2.  将鼠标指针悬停在顶部的`SUPPORT`选项卡上。
3.  单击**代码下载**和**勘误表**。
4.  在**搜索**框中输入书籍的名称。
5.  选择您要下载其代码文件的书。
6.  从购买本书的下拉菜单中选择。
7.  点击**代码下载**。

您还可以通过在 Packt Publishing 网站上的图书网页上单击**代码文件**按钮来下载代码文件。 通过在**搜索**框中输入书名可以访问此页面。 请注意，您需要登录到 Packt 帐户。

下载文件后，请确保使用以下最新版本解压缩或解压缩文件夹：

*   Windows 的 WinRAR/7-Zip
*   适用于 Mac 的 Zipeg/iZip/UnRarX
*   适用于 Linux 的 7-Zip/PeaZip

本书的代码包也托管在 [GitHub](https://github.com/TrainingByPackt/Beginning-Application-Developmentwith-TensorFlow-and-Keras) 上。 我们还从[这里](https://github.com/PacktPublishing/)提供了丰富的书籍和视频目录中的其他代码包。 去看一下！

# 安装

在开始本课程之前，我们将安装 Visual Studio Code，Python 3，TensorFlow 和 Keras。 安装步骤如下：

## 安装 Visual Studio

1.  在浏览器中访问[这里](https://code.visualstudio.com/)。
2.  单击主页右上角的下载。
3.  接下来，选择 Windows。
4.  按照安装程序中的步骤就可以了！ 您的 Visual Studio Code 已准备就绪。

## 安装 Python 3

1.  转到[这里](https://www.python.org/downloads/)。
2.  单击 Download Python 3.6.4 选项下载设置。
3.  按照安装程序中的步骤就可以了！ 您的 Python 已准备就绪。

## 安装 TensorFlow

按照[以下网站](https://www.tensorflow.org/install/install_windows)上的说明下载并安装 TensorFlow。

## 安装 Keras

按照[以下网站](https://keras.io/#installation)上的说明下载并安装 Keras。