# 零、前言

TensorFlow 是 Google 流行的机器学习和深度学习产品。 它已迅速成为执行快速，高效和准确的深度学习任务的工具的流行选择。

这本书向您展示了现实世界项目的实际实现，并教您如何利用 TensorFlow 的功能执行高效的深度学习。 在本书中，您将了解执行深度学习的不同范例，例如深度神经网络，卷积神经网络，循环神经网络等等，以及如何使用 TensorFlow 实现它们。

这将在自然语言处理，图像分类和欺诈检测等热门话题领域的三个真实项目的端到端实现的帮助下进行演示。

到本书结尾，您将掌握 TensorFlow 和 Keras 的所有深度学习概念及其实现。

# 这本书是给谁的

本书面向希望将机器学习集成到应用软件中并通过在 TensorFlow 中实现实际项目来掌握深度学习的应用开发人员，数据科学家和机器学习从业人员。 要充分利用本书，需要具备 Python 编程知识和深度学习的基础知识。

# 本书涵盖的内容

第 1 章，“机器学习工具包”着眼于安装 Docker，设置机器学习 Docker 文件，与主机共享数据并运行 REST 服务以提供环境。

第 2 章，“图像数据”，教 MNIST 数字，如何获取它们，张量实际上只是多维数组，以及如何将图像数据和分类或分类数据编码为张量 。 然后，我们进行了快速回顾并采用了秘籍的方法来考虑大小和张量，以便为机器学习准备数据。

第 3 章，“经典神经网络”涵盖了很多内容！ 我们看到了经典的或密集的神经网络的结构。 我们了解激活，非线性和 softmax。 然后我们建立测试和训练数据，并学习如何使用`Dropout`和`Flatten`构建网络。 我们还将学习有关求解器的所有知识，或者机器实际学习的方法。 然后，我们探索超参数，最后，通过网格搜索对模型进行微调。

第 4 章，“卷积神经网络”教您卷积，卷积是一种在图像上移动以提取特征的松散连接方式。 然后，我们学习有关池化的知识，该池总结了最重要的特征。 我们将使用这些技术构建卷积神经网络，并将卷积和池化的许多层结合起来，以生成一个深度神经网络。

第 5 章，“图像分类服务器”使用 Swagger API 定义创建 REST API 模型，然后声明性地生成 Python 框架，以便我们为该 API 提供服务。 然后，我们创建一个 Docker 容器，该容器不仅捕获我们正在运行的代码（即我们的服务），而且还捕获我们的预训练机器学习模型。 然后形成一个包，以便我们能够部署和使用我们的容器。 最后，我们使用此容器提供服务并做出预测。

# 充分利用这本书

你需要：

*   命令行 Shell 经验
*   有 Python 脚本编写或应用开发经验

# 下载示例代码文件

您可以从 [www.packtpub.com](http://www.packtpub.com) 的帐户中下载本书的示例代码文件。 如果您在其他地方购买了此书，则可以访问 [www.packtpub.com/support](http://www.packtpub.com/support) 并注册以将文件直接通过电子邮件发送给您。

您可以按照以下步骤下载代码文件：

1.  登录或注册 [www.packtpub.com](http://www.packtpub.com/support) 。
2.  选择支持选项卡。
3.  单击代码下载和勘误。
4.  在搜索框中输入书籍的名称，然后按照屏幕上的说明进行操作。

下载文件后，请确保使用以下最新版本解压缩或解压缩文件夹：

*   Windows 的 WinRAR/7-Zip
*   Mac 版 Zipeg/iZip/UnRarX
*   适用于 Linux 的 7-Zip/PeaZip

本书的代码包也托管在 [GitHub](https://github.com/PacktPublishing/Hands-On-Deep-Learning-for-Images-with-TensorFlow)。 如果代码有更新，它将在现有的 GitHub 存储库中进行更新。

我们还有丰富的书籍和视频目录中的其他代码包，可通过[这里](https://github.com/PacktPublishing/)获得。 去看一下！

# 使用约定

本书中使用了许多文本约定。

`CodeInText`：指示文本，数据库表名称，文件夹名称，文件名，文件扩展名，路径名，虚拟 URL，用户输入和 Twitter 句柄中的代码字。 这是一个示例：“您只需键入`docker --help`以确保已安装所有组件。”

任何命令行输入或输出的编写方式如下：

```py
C:\11519>docker build -t keras .
```

**粗体**：表示新术语，重要单词或您在屏幕上看到的单词。 例如，菜单或对话框中的单词会出现在这样的文本中。 这是一个示例：“我们将选择并复制稍后将使用的测试命令，然后单击“应用”。

警告或重要提示如下所示。

提示和技巧如下所示。