# 零、前言

Python 是一种编程语言，可提供数据科学领域的各种功能。 在本书中，我们将介绍两个 Python 库，即 scikit-learn 和 TensorFlow。 我们将学习集成方法的各种实现，如何将它们与实际数据集结合使用，以及它们如何提高分类和回归问题中的预测准确率。

本书从学习集成方法及其功能开始。 我们将研究 scikit-learn 如何提供正确的工具来选择模型的超参数。 从那里开始，我们将深入到预测分析的本质，并探索其各种特征。 我们将向您介绍人工神经网络，TensorFlow，以及用于构建神经网络的核心概念。

在最后一节中，我们将考虑诸如计算能力，改进的方法以及软件改进等因素，以进行有效的预测分析。 您将精通使用 DNN 解决常见的挑战。

# 这本书是给谁的

本书适用于对使用 Python 实现高级预测分析感兴趣的数据分析师，软件工程师和机器学习开发人员。 商业智能专家还将发现本书必不可少，因为它将教给他们如何从基本的预测模型转变为建立高级模型并产生更好的预测。 假定您具有 Python 知识并熟悉预测分析概念。

# 本书涵盖的内容

第 1 章，“回归和分类的集成方法”涵盖了集成方法或算法的应用，以产生模型的准确预测。 我们将通过集成方法来解决回归和分类问题。

第 2 章，“交叉验证和参数调整”探索了各种技术来组合和构建更好的模型。 我们将学习不同的交叉验证方法，包括保留交叉验证和 k 折交叉验证。 我们还将讨论什么是超参数调优。

第 3 章，“处理特征”，探讨了特征选择方法，降维，PCA 和特征工程。 我们还将研究通过特征工程改进模型的方法。

第 4 章，“人工神经网络和 TensorFlow 简介”是 ANN 和 TensorFlow 的简介。 我们将探索网络中的各种元素及其功能。 我们还将在其中学习 TensorFlow 的基本概念。

第 5 章，“使用 TensorFlow 和深度神经网络进行预测分析”，在 TensorFlow 和深度学习的帮助下探索预测分析。 我们将研究 MNIST 数据集和使用该数据集的模型分类。 我们将学习 DNN，它们的功能以及 DNN 在 MNIST 数据集中的应用。

# 充分利用这本书

本书介绍了一些最先进的预测分析工具，模型和技术。 主要目标是向观众展示如何提高预测模型的表现，首先，通过展示如何构建更复杂的模型，其次，通过展示如何使用相关技术来显着提高预测模型的质量。

# 下载示例代码文件

您可以从 [www.packt.com](http://www.packt.com) 的帐户中下载本书的示例代码文件。 如果您在其他地方购买了此书，则可以访问 [www.packt.com/support](http://www.packt.com/support) 并注册以将文件直接通过电子邮件发送给您。

您可以按照以下步骤下载代码文件：

1.  登录或注册 [www.packt.com](http://www.packt.com) 。
2.  选择支持选项卡。
3.  单击代码下载和勘误。
4.  在搜索框中输入书籍的名称，然后按照屏幕上的说明进行操作。

下载文件后，请确保使用以下最新版本解压缩或解压缩文件夹：

*   Windows 的 WinRAR/7-Zip
*   Mac 版 Zipeg/iZip/UnRarX
*   适用于 Linux 的 7-Zip/PeaZip

本书的代码包也托管在 [GitHub](https://github.com/PacktPublishing) 上。 如果代码有更新，它将在现有的 GitHub 存储库中进行更新。

我们还有丰富的书籍和视频目录中的其他代码包，可通过[这里](https://github.com/PacktPublishing/)获得。 去看一下！

# 下载彩色图像

我们还提供了 PDF 文件，其中包含本书中使用的屏幕截图/图表的彩色图像。 [您可以在此处下载](https://www.packtpub.com/sites/default/files/downloads/9781789617740_ColorImages.pdf)。

# 使用约定

本书中使用了许多文本约定。

`CodeInText`：指示文本，数据库表名称，文件夹名称，文件名，文件扩展名，路径名，虚拟 URL，用户输入和 Twitter 句柄中的代码字。 这是一个示例：“以下屏幕截图显示了用于导入`train_test_split`函数和`RobustScaler`方法的代码行。”

代码块设置如下：

```py
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
%matplotlib inline
```

**粗体**：表示新术语，重要单词或您在屏幕上看到的单词。 例如，菜单或对话框中的单词会出现在这样的文本中。 这是一个示例：“用于为特定数据集选择最佳估计器或为所有超参数选择最佳值的方法称为**超参数调优**。”

警告或重要提示如下所示。

提示和技巧如下所示。